/************************************************************************
 *                                                                      *
 * File: spvkxb.c                                                       *
 * Example problem, with sensitivity analysis                           *
 *----------------------------------------------------------------------*
 * An ODE system is generated from the following 2-species diurnal      *
 * kinetics advection-diffusion PDE system in 2 space dimensions:       *
 *                                                                      *
 * dc(i)/dt = Kh*(d/dx)^2 c(i) + V*dc(i)/dx + (d/dy)(Kv(y)*dc(i)/dy)    *
 *                 + Ri(c1,c2,t)      for i = 1,2,   where              *
 *   R1(c1,c2,t) = -q1*c1*c3 - q2*c1*c2 + 2*q3(t)*c3 + q4(t)*c2 ,       *
 *   R2(c1,c2,t) =  q1*c1*c3 - q2*c1*c2 - q4(t)*c2 ,                    *
 *   Kv(y) = Kv0*exp(y/5) ,                                             *
 * Kh, V, Kv0, q1, q2, and c3 are constants, and q3(t) and q4(t)        *
 * vary diurnally.   The problem is posed on the square                 *
 *   0 <= x <= 20,    30 <= y <= 50   (all in km),                      *
 * with homogeneous Neumann boundary conditions, and for time t in      *
 *   0 <= t <= 86400 sec (1 day).                                       *
 * The PDE system is treated by central differences on a uniform        *
 * mesh, with simple polynomial initial profiles.                       *
 *                                                                      *
 * The problem is solved by PVODE on NPE processors, treated as a       *
 * rectangular process grid of size NPEX by NPEY, with NPE = NPEX*NPEY. *
 * Each processor contains a subgrid of size MXSUB by MYSUB of the      *
 * (x,y) mesh.  Thus the actual mesh sizes are MX = MXSUB*NPEX and      *
 * MY = MYSUB*NPEY, and the ODE system size is NY = 2*MX*MY.           *
 *                                                                      *
 * The solution with PVODE is done with the BDF/GMRES method (i.e.      *
 * using the CVSPGMR linear solver) and a block-diagonal matrix with    *
 * banded blocks as a left preconditioner, using the PVBBDPRE module.   *
 * A copy of the approximate Jacobian is saved and conditionally reused *
 * within the Precond routine.                                          *
 *                                                                      *
 * Performance data and sampled solution values are printed at selected *
 * output times, and all performance counters are printed on completion.*
 *                                                                      *
 * This version uses MPI for user routines, and the MPI_PVODE solver.   *   
 * Execute with number of processors = NPEX*NPEY (see constants below). *
 ************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "diurnal_func.h"
#include "llnltyps.h"  /* definitions of real, integer, TRUE, FALSE       */
#include "cvode.h"     /* main CVODE header file                          */
#include "iterativ.h"  /* contains the enum for types of preconditioning  */
#include "cvspgmr.h"   /* use CVSPGMR linear solver                       */
#include "pvbbdpre.h"  /* band preconditioner function prototypes         */
#include "nvector.h"   /* definitions of type N_Vector, macro N_VDATA     */
#include "llnlmath.h"  /* contains SQR macro                              */
#include "mpi.h"       /* MPI data types and prototypes                   */

#include "sensitivity.h"

/* CVodeMalloc Constants */

#define RTOL    1.0e-5            /* scalar relative tolerance */
#define FLOOR   100.0             /* value of C1 or C2 at which tolerances */
                                  /* change from relative to absolute      */
#define ATOL    (RTOL*FLOOR)      /* scalar absolute tolerance */


/* Prototypes of private helper functions */

static void InitUserData(integer my_pe, MPI_Comm comm, UserData data);
static void SetInitialProfiles(N_Vector u, UserData data);
static void PrintOutput(integer my_pe, MPI_Comm comm, long int iopt[],
                        real ropt[], N_Vector u, real t);
static void PrintFinalStats(long int iopt[]);

/* Prototype of function called by the CVODE solver */

void f(integer N, nosens_real t, N_Vector u, N_Vector udot, void *f_data);


/***************************** Main Program ******************************/

main(int argc, char *argv[])
{
  real abstol, reltol, t, tout, ropt[OPT_SIZE];
  long int iopt[OPT_SIZE];
  N_Vector u;
  UserData data;
  PVBBDData pdata;
  void *cvode_mem;
  int iout, flag, i;
  integer local_N, my_pe, npes, ml, mu;
  machEnvType machEnv;
  MPI_Comm comm;
  real Q1, Q2, C3, A3, A4, KH, VEL, KV0;
  N_Vector *usub;
  integer Ntotal;
  double startwtime, endwtime;

  /* Set problem size */
  Ntotal = (1+NS)*NY;

  /* Set problem parameters */
  Q1 = 1.63e-16;
  Q2 = 4.66e-16;
  C3 = 3.7e16;
  A3 = 22.62;
  A4 = 7.601;
  KH = 4.0e-6;
  VEL = 0.001;
  KV0 = 1.0e-8;

  /* Get processor number and total number of pe's */
  ad_AD_Init(NS);
  ad_MPI_Init(&argc, &argv);
  comm = MPI_COMM_WORLD;
  MPI_Comm_size(comm, &npes);
  MPI_Comm_rank(comm, &my_pe);

  if (npes != NPEX*NPEY) {
    if (my_pe == 0)
      printf("\n npes=%d is not equal to NPEX*NPEY=%d\n", npes,NPEX*NPEY);
    return(1);
  }

  /* Set local length */
  local_N = NVARS*MXSUB*MYSUB;

  /* Set machEnv block */
  machEnv = PVecInitMPI(comm, local_N, NY, &argc, &argv);
  if (machEnv == NULL) return(1);

  /* Allocate and load user data block */
  data = (UserData) malloc(sizeof *data);
  data->p = (real *) malloc(NP*sizeof(real));
  data->p[0] = Q1;
  data->p[1] = Q2;
  data->p[2] = C3;
  data->p[3] = A3;
  data->p[4] = A4;
  data->p[5] = KH;
  data->p[6] = VEL;
  data->p[7] = KV0;
  InitUserData(my_pe, comm, data);

  /* Scaling factor for each sensitivity equation */
  data->pbar = (real *) malloc(NP*sizeof(real));
  for (i = 1; i <= NS; i++)
    data->pbar[i-1] = data->p[i-1];

  /* Allocate preconditioner block */
  mu = ml = 2;
  pdata = PVBBDAlloc(local_N, mu, ml, 0.0, flocal, ucomm, data);
  if (pdata == NULL) { printf("PVBBDAlloc failed."); return(1); }

  /* Allocate u, and set initial values and tolerances */ 
  u = N_VNew(Ntotal, machEnv);
  SetInitialProfiles(u, data);
  abstol = ATOL; reltol = RTOL;

  /* Sensitivity and state variables have same scalar error  tolerances */

  /* Set initial values for sensitivity variables */
  usub = (N_Vector *) u->subptr;
  for (i = 1; i <= NS; i++)
    N_VConst(ZERO, usub[i]);

  /* Set (optional) inputs in iopt and ropt arrays */
  for (i=0; i < OPT_SIZE; i++) {
    iopt[i] = 0;
    ropt[i] = 0.0;
  }
  iopt[MXSTEP] = 10000;

/* Call SensCVodeMalloc to initialize CVODE: 

     Ntotal  is the problem size = total number of ODEs
     f       is the user's right hand side function in u'=f(t,u)
     T0      is the initial time
     u       is the initial dependent variable vector
     BDF     specifies the Backward Differentiation Formula
     NEWTON  specifies a Newton iteration
     SS      specifies scalar relative and absolute tolerances
     &reltol and &abstol are pointers to the scalar tolerances
     data    is the pointer to the user-defined data block
     NULL    is the pointer to the error message file
     TRUE    indicates there are optional inputs in iopt and ropt
     iopt, ropt  communicate optional integer and real input/output

     A pointer to CVODE problem memory is returned and stored in cvode_mem.  */

  cvode_mem = SensCVodeMalloc(Ntotal, f, T0, u, BDF, NEWTON, SS, &reltol,
                          &abstol, data, NULL, TRUE, iopt, ropt, machEnv,
			  NY, NS, data->p, data->pbar, RHOMAX);

  if (cvode_mem == NULL) { printf("SensCVodeMalloc failed."); return(1); }

  /* Call SensCVSpgmr to specify the CVODE linear solver CVSPGMR with
     left preconditioning, modified Gram-Schmidt orthogonalization,
     default values for the maximum Krylov dimension maxl and the tolerance
     parameter delt, preconditioner setup and solve routines from the
     PVBBDPRE module, and the pointer to the preconditioner data block.    */

  SensCVSpgmr(cvode_mem, LEFT, MODIFIED_GS, 0, 0.0, PVBBDPrecon, PVBBDPSol, pdata);

  /*
  if (my_pe == 0) {
    printf("\n2-species diurnal advection-diffusion problem\n");
    printf("  %d by %d mesh on %d processors\n",MX,MY,npes);
    printf("  Using PVBBDPRE preconditioner module, mu = %d, ml = %d\n\n",
           mu, ml); 
  }
  */

  /* Pre-load to make timings more accurate */

  flag = CVode(cvode_mem, TWOHR, u, &t, NORMAL);
  if (flag != SUCCESS) {
    if (my_pe == 0) printf("CVode failed, flag =%d.\n", flag);
    return;
  }

  startwtime = MPI_Wtime();

  /* In loop over output points, call CVode, print results, test for error */

  for (iout=2, tout = 2*TWOHR; iout <= NOUT; iout++, tout += TWOHR) {
    flag = CVode(cvode_mem, tout, u, &t, NORMAL);
    /* PrintOutput(my_pe, comm, iopt, ropt, u, t); */
    if (flag != SUCCESS) {
      if (my_pe == 0) printf("CVode failed, flag =%d.\n", flag);
      break;
    }
  }

  endwtime = MPI_Wtime();

  /* Print final statistics and free memory */

  if (my_pe == 0) {
    /* N_VPrint(u); */

    PrintFinalStats(iopt);
    printf("In PVBBDPRE: real/integer local work space sizes = %d, %d\n",
           PVBBD_RPWSIZE(pdata), PVBBD_IPWSIZE(pdata) );  
    printf("             no. flocal evals. = %d\n",PVBBD_NGE(pdata));
    printf("Time = %g\n",endwtime-startwtime);
  }
  SensCVodeFree(cvode_mem);
  N_VFree(u);
  free(data->pbar);
  free(data->p);
  free(data);

  PVBBDFree(pdata);
  PVecFreeMPI(machEnv);
  MPI_Finalize();
  return(0);
}


/*********************** Private Helper Functions ************************/

/* Load constants in data */

static void InitUserData(integer my_pe, MPI_Comm comm, UserData data)
{
  integer isubx, isuby;
  real KH, VEL, KV0;

  /* Load problem coefficients and parameters */
  KH = data->p[5];
  VEL = data->p[6];
  KV0 = data->p[7];

  /* Set problem constants */
  data->om = PI/HALFDAY;
  data->dx = (XMAX-XMIN)/((real)(MX-1));
  data->dy = (YMAX-YMIN)/((real)(MY-1));

  /* Set machine-related constants */
  data->comm = comm;
  data->my_pe = my_pe;
  /* isubx and isuby are the PE grid indices corresponding to my_pe */
  isuby = my_pe/NPEX;
  isubx = my_pe - isuby*NPEX;
  data->isubx = isubx;
  data->isuby = isuby;
  /* Set the sizes of a boundary x-line in u and uext */
  data->nvmxsub = NVARS*MXSUB;
  data->nvmxsub2 = NVARS*(MXSUB+2);

}

/* Set initial conditions in u */

static void SetInitialProfiles(N_Vector u, UserData data)
{
  integer isubx, isuby, lx, ly, jx, jy, offset;
  real dx, dy, x, y, cx, cy, xmid, ymid;
  real *uarray;

  /* Set pointer to data array in vector u */

  uarray = N_VDATA(u);

  /* Get mesh spacings, and subgrid indices for this PE */

  dx = data->dx;         dy = data->dy;
  isubx = data->isubx;   isuby = data->isuby;

  /* Load initial profiles of c1 and c2 into local u vector.
  Here lx and ly are local mesh point indices on the local subgrid,
  and jx and jy are the global mesh point indices. */

  offset = 0;
  xmid = .5*(XMIN + XMAX);
  ymid = .5*(YMIN + YMAX);
  for (ly = 0; ly < MYSUB; ly++) {
    jy = ly + isuby*MYSUB;
    y = YMIN + jy*dy;
    cy = SQR(0.1*(y - ymid));
    cy = 1.0 - cy + 0.5*SQR(cy);
    for (lx = 0; lx < MXSUB; lx++) {
      jx = lx + isubx*MXSUB;
      x = XMIN + jx*dx;
      cx = SQR(0.1*(x - xmid));
      cx = 1.0 - cx + 0.5*SQR(cx);
      uarray[offset  ] = C1_SCALE*cx*cy; 
      uarray[offset+1] = C2_SCALE*cx*cy;
      offset = offset + 2;
    }
  }
}

/* Print current t, step count, order, stepsize, and sampled c1,c2 values */

static void PrintOutput(integer my_pe, MPI_Comm comm, long int iopt[], 
                        real ropt[], N_Vector u, real t)
{
  real *uarray, tempu[2];
  integer npelast, i0, i1;
  MPI_Status status;

  npelast = NPEX*NPEY - 1;
  uarray = N_VDATA(u);

  /* Send c1,c2 at top right mesh point to PE 0 */
  if (my_pe == npelast) {
    i0 = NVARS*MXSUB*MYSUB - 2;
    i1 = i0 + 1;
    if (npelast != 0)
      MPI_Send(&uarray[i0], 2, PVEC_REAL_MPI_TYPE, 0, 0, comm);
    else {
      tempu[0] = uarray[i0];
      tempu[1] = uarray[i1];
    }
  }

  /* On PE 0, receive c1,c2 at top right, then print performance data
     and sampled solution values */ 
  if (my_pe == 0) {
    if (npelast != 0)
      MPI_Recv(&tempu[0], 2, PVEC_REAL_MPI_TYPE, npelast, 0, comm, &status);
    printf("t = %.2e   no. steps = %d   order = %d   stepsize = %.2e\n",
           t, iopt[NST], iopt[QU], ropt[HU]);
    printf("At bottom left:  c1, c2 = %12.3e %12.3e \n", uarray[0], uarray[1]);
    printf("At top right:    c1, c2 = %12.3e %12.3e \n\n", tempu[0], tempu[1]);
  }
}

/* Print final statistics contained in iopt */

static void PrintFinalStats(long int iopt[])
{
  printf("\nFinal Statistics.. \n\n");
  printf("lenrw   = %5ld    leniw = %5ld\n", iopt[LENRW], iopt[LENIW]);
  printf("llrw    = %5ld    lliw  = %5ld\n", iopt[SPGMR_LRW], iopt[SPGMR_LIW]);
  printf("nst     = %5ld    nfe   = %5ld\n", iopt[NST], iopt[NFE]);
  printf("nni     = %5ld    nli   = %5ld\n", iopt[NNI], iopt[SPGMR_NLI]);
  printf("nsetups = %5ld    netf  = %5ld\n", iopt[NSETUPS], iopt[NETF]);
  printf("npe     = %5ld    nps   = %5ld\n", iopt[SPGMR_NPE], iopt[SPGMR_NPS]);
  printf("ncfn    = %5ld    ncfl  = %5ld\n \n", iopt[NCFN], iopt[SPGMR_NCFL]);
}
