/*$Id: ex5d.c,v 1.8 2000/09/22 20:46:14 bsmith Exp $*/

/* Program usage:  mpirun -np <procs> ex5d [-help] [all PETSc options] */

static char help[] = "Solves a nonlinear system in parallel with SNES.\n\
We solve the  Bratu (SFI - solid fuel ignition) problem in a 2D rectangular\n\
domain, using distributed arrays (DAs) to partition the parallel grid.\n\
The command line options include:\n\
  -par <parameter>, where <parameter> indicates the problem's nonlinearity\n\
     problem SFI:  <parameter> = Bratu parameter (0 <= par <= 6.81)\n\
  -mx <xg>, where <xg> = number of grid points in the x-direction\n\
  -my <yg>, where <yg> = number of grid points in the y-direction\n\
  -Nx <npx>, where <npx> = number of processors in the x-direction\n\
  -Ny <npy>, where <npy> = number of processors in the y-direction\n\n";

/*T
   Concepts: SNES^parallel Bratu example
   Concepts: DA^using distributed arrays;
   Processors: n
T*/

/* ------------------------------------------------------------------------

    Solid Fuel Ignition (SFI) problem.  This problem is modeled by
    the partial differential equation
  
            -Laplacian u - lambda*exp(u) = 0,  0 < x,y < 1,
  
    with boundary conditions
   
             u = 0  for  x = 0, x = 1, y = 0, y = 1.
  
    A finite difference approximation with the usual 5-point stencil
    is used to discretize the boundary value problem to obtain a nonlinear 
    system of equations.

    The uniprocessor version of this code is snes/examples/tutorials/ex4.c

    Same as ex5.c except it uses DAVecGetArray() instead of VecGetArray()

  ------------------------------------------------------------------------- */

/* 
   Include "petscda.h" so that we can use distributed arrays (DAs).
   Include "petscsnes.h" so that we can use SNES solvers.  Note that this
   file automatically includes:
     petsc.h       - base PETSc routines   petscvec.h - vectors
     petscsys.h    - system routines       petscmat.h - matrices
     petscis.h     - index sets            petscksp.h - Krylov subspace methods
     petscviewer.h - viewers               petscpc.h  - preconditioners
     petscsles.h   - linear solvers
*/
#include "petscda.h"
#include "petscsnes.h"
#include "ad_deriv.h"

/* 
   User-defined application context - contains data needed by the 
   application-provided call-back routines, FormJacobian() and
   FormFunction().
*/
typedef struct {
   double      param;          /* test problem parameter */
   int         mx,my;          /* discretization in x, y directions */
   Vec         localX,localV,localF; /* ghosted local vectors */
   DA          da;             /* distributed array data structure */
   SNES        snes;           
   int         rank;           /* processor rank */
} AppCtx;

/* 
   User-defined routines
*/
extern int FormFunction(SNES,Vec,Vec,void*),FormInitialGuess(AppCtx*,Vec);
extern int FormJacobian(SNES,Vec,Mat*,Mat*,MatStructure*,void*);
extern int FormJacobianVector(Mat,Vec,Vec);
extern int FormJacobianVectorAD(Mat,Vec,Vec);
extern int localfunction2d(Scalar **,Scalar **,int,int,int,int,void *);

/* AD-generated routine */

extern int ad_localfunction2d(DERIV_TYPE **,DERIV_TYPE **,int,int,int,int,void *);

Mat my_GlobalB;

#undef __FUNC__
#define __FUNC__ "main"
int main(int argc,char **argv)
{
  Vec                    x,r;                 /* solution, residual vectors */
  Mat                    J,B,ADShell;                    /* Jacobian matrix */
  AppCtx                 user;                 /* user-defined work context */
  ISLocalToGlobalMapping isltog;                /* mapping from local-to-global indices */
  int                    its;                  /* iterations for convergence */
  int                    Nx,Ny;               /* number of preocessors in x- and y- directions */
  int                    size;                 /* number of processors */
  int                    m,N,ierr;
  double                 bratu_lambda_max = 6.81,bratu_lambda_min = 0.,fnorm;
  PetscTruth             matrix_free_ad;
  DALocalInfo            dainfo;

  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
     Initialize program
     - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

  PetscInitialize(&argc,&argv,(char *)0,help);
  ierr = MPI_Comm_rank(PETSC_COMM_WORLD,&user.rank);CHKERRQ(ierr);

  /*
     Initialize problem parameters
  */
  user.mx = 4; user.my = 4; user.param = 6.0;
  ierr = PetscOptionsGetInt(PETSC_NULL,"-mx",&user.mx,PETSC_NULL);CHKERRQ(ierr);
  ierr = PetscOptionsGetInt(PETSC_NULL,"-my",&user.my,PETSC_NULL);CHKERRQ(ierr);
  ierr = PetscOptionsGetDouble(PETSC_NULL,"-par",&user.param,PETSC_NULL);CHKERRQ(ierr); 
  ierr = PetscOptionsHasName(PETSC_NULL,"-snes_mfad",&matrix_free_ad);CHKERRQ(ierr);

  if (user.param >= bratu_lambda_max || user.param <= bratu_lambda_min) {
    SETERRQ(0,"Lambda is out of range");
  }
  N = user.mx*user.my;

  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
     Create nonlinear solver context
     - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

  ierr = SNESCreate(PETSC_COMM_WORLD,SNES_NONLINEAR_EQUATIONS,&user.snes);CHKERRQ(ierr);

  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
     Create vector data structures; set function evaluation routine
     - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

  /*
     Create distributed array (DA) to manage parallel grid and vectors
  */
  ierr = MPI_Comm_size(PETSC_COMM_WORLD,&size);CHKERRQ(ierr);
  Nx = PETSC_DECIDE; Ny = PETSC_DECIDE;
  ierr = PetscOptionsGetInt(PETSC_NULL,"-Nx",&Nx,PETSC_NULL);CHKERRQ(ierr);
  ierr = PetscOptionsGetInt(PETSC_NULL,"-Ny",&Ny,PETSC_NULL);CHKERRQ(ierr);
  if (Nx*Ny != size && (Nx != PETSC_DECIDE || Ny != PETSC_DECIDE))
    SETERRQ(0,"Incompatible number of processors:  Nx * Ny != size");
  ierr = DACreate2d(PETSC_COMM_WORLD,DA_NONPERIODIC,DA_STENCIL_STAR,user.mx,
                    user.my,Nx,Ny,1,1,PETSC_NULL,PETSC_NULL,&user.da);CHKERRQ(ierr);

  /*
     Extract global and local vectors from DA; then duplicate for remaining
     vectors that are the same types
  */
  ierr = DACreateGlobalVector(user.da,&x);CHKERRQ(ierr);
  ierr = VecDuplicate(x,&r);CHKERRQ(ierr);

  ierr = DACreateLocalVector(user.da,&user.localX);CHKERRQ(ierr);
  ierr = VecDuplicate(user.localX,&user.localF);CHKERRQ(ierr);
  ierr = VecDuplicate(user.localX,&user.localV);CHKERRQ(ierr);

  /* 
     Set function evaluation routine and vector
  */
  ierr = SNESSetFunction(user.snes,r,FormFunction,(void*)&user);CHKERRQ(ierr);

  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
     Create matrix data structure; set Jacobian evaluation routine
     - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
  
  ierr = VecGetLocalSize(x,&m);CHKERRQ(ierr);

  ierr = MatCreateMPIAIJ(PETSC_COMM_WORLD,m,m,N,N,5,PETSC_NULL,3,PETSC_NULL,&B);CHKERRQ(ierr);


  if (matrix_free_ad) {
    int localsize,globalsize;

    ierr = DAGetLocalInfo(user.da,&dainfo); CHKERRQ(ierr);
    localsize = dainfo.xm * dainfo.ym * dainfo.zm * dainfo.dof;
    globalsize = dainfo.mx * dainfo.my * dainfo.mz * dainfo.dof;
    ierr = MatCreateShell(PETSC_COMM_WORLD,localsize,localsize,globalsize,globalsize,&user,&ADShell); CHKERRQ(ierr);
    MatShellSetOperation(ADShell,MATOP_MULT,(void(*)())FormJacobianVectorAD);
    J = ADShell;
  } else {
    J = B;
  }

  ierr = SNESSetJacobian(user.snes,J,B,FormJacobian,&user);CHKERRQ(ierr);

    /*
       Get the mapping from local-to-global node numbers for all local nodes,
       including ghost points.  Associate this mapping with the matrix for later
       use in setting matrix entries via MatSetValuesLocal().
    */
  ierr = DAGetISLocalToGlobalMapping(user.da,&isltog);CHKERRQ(ierr);
  ierr = MatSetLocalToGlobalMapping(B,isltog);CHKERRQ(ierr);
  
  my_GlobalB = B;

  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
     Customize nonlinear solver; set runtime options
   - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

  /*
     Set runtime options (e.g., -snes_monitor -snes_rtol <rtol> -ksp_type <type>)
  */
ierr = SNESSetFromOptions(user.snes);CHKERRQ(ierr);

  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
     Evaluate initial guess; then solve nonlinear system
   - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
  /*
     Note: The user should initialize the vector, x, with the initial guess
     for the nonlinear solver prior to calling SNESSolve().  In particular,
     to employ an initial guess of zero, the user should explicitly set
     this vector to zero by calling VecSet().
  */
ierr = FormInitialGuess(&user,x);CHKERRQ(ierr);
ierr = SNESSolve(user.snes,x,&its);CHKERRQ(ierr); 
ierr = FormFunction(user.snes,x,r,(void *)&user);CHKERRQ(ierr);
ierr = VecNorm(r,NORM_2,&fnorm);CHKERRQ(ierr); 
ierr = PetscPrintf(PETSC_COMM_WORLD,"Number of Newton iterations = %d fnorm %g\n",its,fnorm);CHKERRQ(ierr);

  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
     Free work space.  All PETSc objects should be destroyed when they
     are no longer needed.
   - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

ierr = MatDestroy(J);CHKERRQ(ierr);
  ierr = VecDestroy(user.localX);CHKERRQ(ierr); ierr = VecDestroy(x);CHKERRQ(ierr);
  ierr = VecDestroy(user.localF);CHKERRQ(ierr); ierr = VecDestroy(r);CHKERRQ(ierr);      
  ierr = SNESDestroy(user.snes);CHKERRQ(ierr);  ierr = DADestroy(user.da);CHKERRQ(ierr);
  PetscFinalize();

  return 0;
}
/* ------------------------------------------------------------------- */
#undef __FUNC__
#define __FUNC__ "FormInitialGuess"
/* 
   FormInitialGuess - Forms initial approximation.

   Input Parameters:
   user - user-defined application context
   X - vector

   Output Parameter:
   X - vector
 */
int FormInitialGuess(AppCtx *user,Vec X)
{
  int     i,j,mx,my,ierr,xs,ys,xm,ym,gxm,gym,gxs,gys;
  double  one = 1.0,lambda,temp1,temp,hx,hy;
  Scalar  **x;
  Vec     localX = user->localX;

  mx = user->mx;            my = user->my;            lambda = user->param;
  hx = one/(double)(mx-1);  hy = one/(double)(my-1);
  temp1 = lambda/(lambda + one);

  /*
     Get local grid boundaries (for 2-dimensional DA):
       xs, ys   - starting grid indices (no ghost points)
       xm, ym   - widths of local grid (no ghost points)
       gxs, gys - starting grid indices (including ghost points)
       gxm, gym - widths of local grid (including ghost points)
  */
  ierr = DAGetCorners(user->da,&xs,&ys,PETSC_NULL,&xm,&ym,PETSC_NULL);CHKERRQ(ierr);
  ierr = DAGetGhostCorners(user->da,&gxs,&gys,PETSC_NULL,&gxm,&gym,PETSC_NULL);CHKERRQ(ierr);

  /*
     Get a pointer to vector data.
       - For default PETSc vectors, DAVecGetArray() returns a pointer to
         the data array.  Otherwise, the routine is implementation dependent.
       - You MUST call DAVecRestoreArray() when you no longer need access to
         the array.

     The "x" and "y" logical coordinates into x[] are switched because this is C.
     Moving in the "x" coordinate is fastest which means that it is the last coordinate
     instead of the first as it would be in Fortran.
  
  */
  ierr = DAVecGetArray(user->da,localX,(void**)&x);CHKERRQ(ierr);

  /*
     Compute initial guess over the locally owned part of the grid
  */
  for (j=ys; j<ys+ym; j++) {
    temp = (double)(PetscMin(j,my-j-1))*hy;
    for (i=xs; i<xs+xm; i++) {
      if (i == 0 || j == 0 || i == mx-1 || j == my-1) {
        x[j][i] = 0.0; 
      } else {
        x[j][i] = temp1*sqrt(PetscMin((double)(PetscMin(i,mx-i-1))*hx,temp)); 
      }
    }
  }

  /*
     Restore vector
  */
  ierr = DAVecRestoreArray(user->da,localX,(void**)&x);CHKERRQ(ierr);

  /*
     Insert values into global vector
  */
  ierr = DALocalToGlobal(user->da,localX,INSERT_VALUES,X);CHKERRQ(ierr);
  return 0;
} 

